/*
 *  ========== Copyright (c) Valve Corporation. All rights reserved. ==========
 */

#pragma once

#include "common/interface.h"
#include "common/platform.h"
#include <cstdint>

namespace rehlds::common::vgui
{
    /**
     * @brief Handle to an internal VGui panel.
     * This is the only handle to a panel that is valid across dll boundaries.
     */
    using VPanel = std::uint32_t;

    //
    // Handles to vgui objects.
    // NULL values signify an invalid value.
    //

    using HScheme = unsigned long;
    using HTexture = unsigned long;
    using HCursor = unsigned long;
    using HPanel = unsigned long;
    using HFont = unsigned long;

    constexpr HPanel INVALID_PANEL = 0xFFFF'FFFFU;
    constexpr HFont INVALID_FONT = 0;
    constexpr auto STEREO_NOOP = 1.F;
    constexpr auto STEREO_INVALID = 0.F;
}

namespace rehlds::common
{
    constexpr auto* INTERFACE_VGUI_MODULE = "VGuiModule002";
    constexpr auto* INTERFACE_VGUI_MODULE_ADMIN_SERVER = "VGuiModuleAdminServer001";

    /**
     * @brief Standard interface to loading VGui modules.
     */
    class NO_VTABLE IVGuiModule : public IBaseInterface
    {
      public:
        /**
         * @brief Called first to setup the module with the VGui.
         *
         * @return \c true on success, \c false on failure.
         */
        virtual bool initialize(CreateInterfaceFn* vgui_factories, int factory_count) = 0;

        /**
         * @brief Called after all the modules have been initialized.
         *
         * @note Modules should use this time to link to all the other module interfaces.
         */
        virtual bool post_initialize(CreateInterfaceFn* modules, int factory_count) = 0;

        /**
         * @brief Called when the module is selected from the menu or otherwise activated.
         */
        virtual bool activate() = 0;

        /**
         * @brief Returns \c true if the module is successfully initialized and available.
         */
        virtual bool is_valid() = 0;

        /**
         * @brief Requests that the UI is temporarily disabled and all data files saved.
         */
        virtual void deactivate() = 0;

        /**
         * @brief Restart from a \c deactivate()
         */
        virtual void reactivate() = 0;

        /**
         * @brief Called when the module is about to be shutdown.
         */
        virtual void shutdown() = 0;

        /**
         * @brief Returns a handle to the main module panel.
         */
        virtual vgui::VPanel get_panel() = 0;

        /**
         * @brief Sets the parent of the main module panel.
         */
        virtual void set_parent(vgui::VPanel parent) = 0;

        // Messages sent through the panel returned by GetPanel():
        //
        // "ConnectedToGame"        "ip"    "port"  "gamedir"
        // "DisconnectedFromGame"
        // "ActiveGameName"         "name"
        // "LoadingStarted"         "type"  "name"
        // "LoadingFinished"        "type"  "name"
    };
}
